{ StrTEXT.pas  v1.0    Copyright 1994 Adolfo Di Mare }

    { Turbo Pascal String Text File Device Driver }

{$I ADT-ADH.inc}

UNIT StrTEXT;
{ RESULTS
  Implements the AssignStr(F,s) function that will let
  a programmer to use the standard TEXT file input and
  output operations in the string variable "s".

  - This is very handy because oftentimes a program
    was written to use input or output from a text
    file, but for some reason it will be usefull to
    the programmer to fake the text file using a
    Turbo Pascal string.

  - One application where StrTEXT comes dandy is to
    write test programs, because the output that
    would otherwise get into a TEXT file comes out in
    a STRING variable, where it can be examined and
    manipulated quite easily. This is why I wrote this
    device driver: to see what was being written out
    in a text FILE without leaving the IDE.

  - Another application is to set up a string with the
    values that need to be read from a text file, and
    then have the program just get them directly from
    the string instead of having to write them to a
    file to later read them. This is quite handy to
    write test cases, because most test cases will fit
    in a string and those strings can be generated
    quite easily in the program. }

{ PROGRAMMER:  ADH ==> Adolfo Di Mare }

{ USAGE
  StrTEXT is a Text File Device Driver. It implements
  the four device driver functions that Turbo uses to
  handle a text file. These functions are not
  interfaced, because the client programmer needs no
  direct access them. According to the Turbo Pascal
  manual, these functions are the following:

    FUNCTION Open  (VAR F : Dos.TextRec) : INTEGER;
    FUNCTION InOut (VAR F : Dos.TextRec) : INTEGER;
    FUNCTION Flush (VAR F : Dos.TextRec) : INTEGER;
    FUNCTION Close (VAR F : Dos.TextRec) : INTEGER;

  To use a STRING variable as as TEXT file, the client
  programmer needs to invoke the AssignStr() operation,
  as it's shown in the following example.
}

{$IFDEF foo} {$IFnDEF foo}  <== neat comment trick

  VAR
    F : TEXT;   { text file used for input/output }
    s : STRING; { string variable }
    ch: CHAR;
  BEGIN
    s := 'Value';             { s = 'Value' }

    StrTEXT.AssignStr(F, s);  { init StrTEXT device }

    { now input/output on "F" goes into "s" }

    { writes 'Value' in OUTPUT }
    Reset(F);
    WHILE NOT Eof(F) DO BEGIN
      Read (F,      ch);      { input comes from "s" }
      Write(OUTPUT, ch);
    END;                    { reads 5 chars: 'Value' }

    ReWrite(F);               { s becomes empty  }

    Write(F, '=', PI:7:5);   { s = '=3.14159'  }

    Flush(F);                 { nothing changes  }

    Write(F,'!');             { s = '=3.14159!' }

    Close(F);                 { s = '=3.14159!' }
  END;

{$ENDIF} {$ENDIF} { foo } { foo }


{ REQUIREMENTS
  As procedure AssignStr(F,s) links text file "F" with
  string  "s", it is incorrect to change the value of
  the string variable "s".

  It is also invalid to change the buffer for F by
  invoking procedure:
    System.SetTextBuf(F, buffer, SizeOf(buffer))
  while this association is still in place.

  The programmer must remember the every WriteLn()
  will append a CR/LF to the output string. Also,
  every invocation to ReadLn() will expect to find
  CR/LF to finish up gobbling up string input. }

{[])===========([]}
{[]} INTERFACE {[]}
{[])===========([]}


PROCEDURE AssignStr(VAR F : TEXT; VAR S:STRING);


{[]================[]}
{|} IMPLEMENTATION {|}
{[]================[]}

{
  This implementation is tricky, because the string
  variable that gets associated to the text file is
  used as the input/output buffer. This is what
  function AssignStr() does. 

  When doing output, Turbo will write into the string
  the value that needs to be written by this device
  driver. This means that the only work  that's left
  to do is to set the string length. However, the
  values for the fields BufPos and BufEnd cannot be
  changed by the output function, as otherwise Turbo
  would overwrite the string in the next output
  function invocation.

  The mission of function StrTEXT.Output() is to set
  the string's length. However, this function is also
  used as the TextRec.FlushFunc, and it performs its
  duty when its called to flush, after Read, Write,
  ReadLn, and WriteLn, because Turbo seldom invokes
  the TextRec.Output function. The reason for this is
  quite simple: the string is the buffer, which means
  that when it fills up is because the device is full,
  and this is an error. The only time when Turbo does
  call the TextRec.Output function is when the string
  device is closed.

  For input the story is a little bit messier, because
  when the file is opened with Reset() then the value
  that will be read is already stored in the buffer,
  because the string is the buffer. This means that
  when Turbo needs another character it will just take
  it from the buffer, until all the characters in the
  buffers are read, and the the string value has been
  exausted. At this time, there will be no more
  characters to read, which means that the Eof()
  condition will be TRUE. But this implies that the
  device driver Input() function only job is to make
  the Eof() condition TRUE.

  All this means that the values for fields BufEnd and
  BufPos need not be changed by the device drivers
  functions.

  For further details, study the section on Text File
  Device Drivers in the Turbo Pascal manual. Also,
  don't forget to look into the "File Types" section
  in the Inside Turbo Pascal chapter, where the layout
  for file variables is documented.

  TextRec.Handle: contains the file's handle that is
    returned by DOS when the file is open. In the case
    of AssignStr(), the value stored in here is always
    $FFFF.

  TextRec.BufPtr: is a pointer to the buffer of
      TextRec.BufSize bytes. For StrTEXT, this points
      fo the first byte in the string associated to
      the file variable (not to the string's length
      byte).

  TextRec.BufPos: is the index of the next character
    in the buffer to read or write (first byte is 0).

  TextRec.BufEnd: is a count of the number of the
    valid characters in the buffer.

  TextRecF.BufSize: total buffer size.

  The condition Eof(F) appears to be equivalent to
    (F.BufPos = F.BufEnd) =?= Eof(F)
}

USES
  Dos;

{$F+} { force FAR calls throughout }

FUNCTION Open  (VAR F: Dos.TextRec): INTEGER; FORWARD;
FUNCTION Input (VAR F: Dos.TextRec): INTEGER; FORWARD;
FUNCTION Output(VAR F: Dos.TextRec): INTEGER; FORWARD;
FUNCTION Ignore(VAR F: Dos.TextRec): INTEGER; FORWARD;

PROCEDURE AssignStr(VAR F : TEXT; VAR s : STRING);
{ RESULTS
  Binds together the text file "F" with string "s".
  - Every input from file "F" will be obtained by
    reading the value stored in string "s".
  - Every output into file "F" will append the output
    value to string "s".

  Instead of reading/writting into a standard text
  file, AssignStr(F,s) redirects the action of the
  standard text file operations so that they read
  or write the value stored in string "s". }

VAR
  name : STRING[8];
BEGIN

  Dos.TextRec(F).Handle    := $FFFF;
  Dos.TextRec(F).Mode      := Dos.fmClosed;

  Dos.TextRec(F).OpenFunc  := @StrTEXT.Open;
  Dos.TextRec(F).CloseFunc := @StrTEXT.Ignore;
  Dos.TextRec(F).FlushFunc := @StrTEXT.Ignore;

  Dos.TextRec(F).BufPtr  := ADDR( s[1] );
  Dos.TextRec(F).BufSize := SizeOf(STRING)-1;

  { make up a nice name for the file }
  name := 'StrTEXT'+ CHAR(0);
  System.Move(
    name[1],
    Dos.TextRec(F).Name,
    Length(name)
  );

END;

TYPE
  TCHARarray = ARRAY[0..SizeOf(STRING)-1] OF CHAR;

FUNCTION Open(VAR F : Dos.TextRec) : INTEGER;
VAR
  p : ^STRING;
  q : ^TCHARarray ABSOLUTE p; { handy to watch "p^" }
BEGIN { Open }
  { "p" points to the string's length byte }
  p := Ptr( Seg(F.BufPtr^), Ofs(F.BufPtr^)-1);

  IF F.Mode = Dos.fmInput THEN BEGIN
    { Reset(F) }
    F.InOutFunc := @StrTEXT.Input;
    F.FlushFunc := @StrTEXT.Ignore;

    { sets up the buffer variables to  }
    { fake that something was read in. }
    F.BufEnd := Length(p^);
    F.BufPos := 0;
  END
  ELSE IF F.Mode = Dos.fmOutput THEN BEGIN
    { ReWrite(F) }
    F.InOutFunc := @StrTEXT.Output;
    F.FlushFunc := @StrTEXT.Output;

    { empty the string by setting its length byte }
    p^[0] := CHAR(0);
  END
  ELSE BEGIN
    { Append(F) }
    F.InOutFunc := @StrTEXT.Output;
    F.FlushFunc := @StrTEXT.Output;

    { tell Turbo that this is a TEXT file }
    F.Mode := Dos.fmOutput;

    F.BufEnd := Length(p^);
    F.BufPos := Length(p^);
  END;

  Open := 0;
END;  { Open }

FUNCTION Input(VAR F : Dos.TextRec) : INTEGER;
BEGIN { Input }

{
  when Input() is invoked it means that the string
  value stored in F.BufPtr^ has been read completely;
  the only work left to do is to make Eof(f)=TRUE
  and return 0.

  This means that BufEnd should not be set to 0,
  because that will make Turbo think that there is
  hope to get further input, and there can be no more
  input because the string was completely read.

  F.BufEnd := 0; NO-NO-NO!
  F.BufPos := 0;

  IF  F.BufEnd <> F.BufPos THEN BEGIN
    WriteLn('This should never happen');
  END;
}

  { can't read: return 0 to make Eof(F)=TRUE }
  Input := 0;
END;  { Input }

FUNCTION Output(VAR F : Dos.TextRec) : INTEGER;
VAR
  p : ^STRING;
  q : ^TCHARarray ABSOLUTE p; { handy to watch "p^" }
BEGIN { Output }
{
  as Turbo already wrote the string value in "s", the
  only work that's left to do is to set its length.
}
  p     := Ptr( Seg(F.BufPtr^), Ofs(F.BufPtr^)-1);
  p^[0] := CHAR(F.BufPos);
{
  can't change F.BufPos, because then Turbo would
  overwrite the string's value.
}
  Output   := 0;
END;  { Output }

FUNCTION Ignore(VAR F : Dos.TextRec) : INTEGER;
BEGIN
  { return success flag, but do nothing }
  Ignore := 0;
END;

BEGIN { StrTEXT }
END.  { StrTEXT }



  Implementa el tipo TStrDev que permite leer y escribir
  en hilearas como si fueran un archivo de Texto (TEXT). }


  Si el archivo de texdo F ha sido est asociado a una
  hilera S por medio de StrTEXT.AssignStr(F, S);
  entonces no debe invocarse al procedimiento
  System.SetTextBuf(F, Buf, Size) mientras esa
  asociacin perdure. }


{ Copyright (c) 1985, 87 by Borland International, Inc. }

unit AuxInOut;
{ AUX example from P-373 of the Owner's Handbook. Use 
  this example instead of the one in the manual.
}

interface

uses Dos;

procedure AssignAux(var F: Text; Port,Params: Word);

implementation

{$R-,S-}

type
  AuxRec = record
             Port,Params: Word;
             Unused: array[1..12] of Byte;
           end;

procedure AuxInit(Port,Params: Word);
inline(
  $58/          { POP   AX      ;Pop parameters         }
  $5A/          { POP   DX      ;Pop port number        }
  $B4/$00/      { MOV   AH,0    ;Code for initialize    }
  $CD/$14);     { INT   14H     ;Call BIOS              }

function AuxInChar(Port: Word): Char;
inline(
  $5A/          { POP   DX      ;Pop port number        }
  $B4/$02/      { MOV   AH,2    ;Code for input         }
  $CD/$14);     { INT   14H     ;Call BIOS              }

procedure AuxOutChar(Port: Word; Ch: Char);
inline(
  $58/          { POP   AX      ;Pop character          }
  $5A/          { POP   DX      ;Pop port number        }
  $B4/$01/      { MOV   AH,1    ;Code for output        }
  $CD/$14);     { INT   14H     ;Call BIOS              }

function AuxInReady(Port: Word): Boolean;
inline(
  $5A/          { POP   DX      ;Pop port number        }
  $B4/$03/      { MOV   AH,3    ;Code for status        }
  $CD/$14/      { INT   14H     ;Call BIOS              }
  $88/$E0/      { MOV   AL,AH   ;Get line status in AH  }
  $24/$01);     { AND   AL,1    ;Isolate Data Ready bit }

{$F+}

function AuxInput(var F: TextRec): Integer;
var
  P: Word;
begin
  with F,AuxRec(UserData) do
  begin
    P:=0;
    while AuxInReady(Port) and (P<BufSize) do
    begin
      BufPtr^[P]:=AuxInChar(Port); Inc(P);
    end;
    BufPos:=0; BufEnd:=P;
  end;
  AuxInput:=0;
end;

function AuxOutput(var F: TextRec): Integer;
var
  P: Word;
begin
  with F,AuxRec(UserData) do
  begin
    P:=0;
    while P<BufPos do
    begin
      AuxOutChar(Port,BufPtr^[P]); Inc(P);
    end;
    BufPos:=0;
  end;
  AuxOutput:=0;
end;

function AuxIgnore(var F: TextRec): Integer;
begin
  AuxIgnore:=0;
end;

function AuxOpen(var F: TextRec): Integer;
begin
  with F,AuxRec(UserData) do
  begin
    AuxInit(Port,Params);
    if Mode=fmInput then
    begin
      InOutFunc:=@AuxInput;
      FlushFunc:=@AuxIgnore;
    end else
    begin
      Mode:=fmOutput;
      InOutFunc:=@AuxOutput;
      FlushFunc:=@AuxOutput;
    end;
    CloseFunc:=@AuxIgnore;
  end;
  AuxOpen:=0;
end;


procedure AssignAux(var F: Text; Port,Params: Word);
begin
  TextRec(F).Handle   := $FFFF;
  TextRec(F).Mode     := fmClosed;

  TextRec(F).BufSize  := SizeOf( (TextRec(F).Buffer) );

  TextRec(F).BufPtr   := @ TextRec(F).Buffer;
  TextRec(F).OpenFunc := @ AuxOpen;

  AuxRec( TextRec(F).UserData ).Port   := Port;
  AuxRec( TextRec(F).UserData ).Params := Params;

  TextRec(F).Name[0] := #0;
end;

end.

  PTextBuf = ^TTextBuf;
  TTextBuf = array[0..127] of Char;
  TTextRec = record
    Handle: Word;
    Mode: Word;

    BufSize: Word;
    Private: Word;

    BufPos: Word;
    BufEnd: Word;

    BufPtr: PTextBuf;

    OpenFunc: Pointer;
    InOutFunc: Pointer;
    FlushFunc: Pointer;
    CloseFunc: Pointer;
    UserData: array[1..16] of Byte;

    Name: array[0..79] of Char;
    Buffer: TTextBuf;
  end;

  TTextRec = record


    BufPos: Word;
    BufEnd: Word;


    InOutFunc: Pointer;


  end;

{ EOF: StrTEXT.pas }
